import re
import os
import errno
import string
import subprocess
import k3color

listtype = (tuple, list)

invisible_chars = "".join(map(chr, list(range(0, 32))))
invisible_chars_re = re.compile("[%s]" % re.escape(invisible_chars))


def break_line(linestr, width):
    lines = linestr.splitlines()
    rst = []

    space = " "
    if isinstance(linestr, k3color.Str):
        space = k3color.Str(" ")

    for line in lines:
        words = line.split(" ")

        buf = words[0]
        for word in words[1:]:
            if len(word) + len(buf) + 1 > width:
                rst.append(buf)
                buf = word
            else:
                buf += space + word

        if buf != "":
            rst.append(buf)

    return rst


def line_pad(linestr, padding=""):
    """

    :param linestr: multiple line string with `\n` as line separator.
    :param padding: left padding string to add before each line.
    It could also be a callable object that returns a string.
    This is useful when creating dynamic padding.

    :return: multiple line string with `\n` as line separator, with left padding added.
    """
    lines = linestr.split("\n")

    if type(padding) in (str, bytes):
        lines = [padding + x for x in lines]

    elif callable(padding):
        lines = [padding(x) + x for x in lines]

    lines = "\n".join(lines)

    return lines


def _to_str(y):
    if isinstance(y, k3color.Str):
        pass
    elif isinstance(y, int):
        y = str(y)
    elif isinstance(y, listtype):
        y = str(y)

    return y


def struct_repr(data, key=None):
    """
    Render primitive or composite data to a structural representation string list.
    :param data: a number, string, list or dict to render to a structural representation.
    :param key: is a callable that is used to sort dict keys. It is used in sort: `keys.sort(key=key)`.
    :return: a list of string.
    Render a data to a multi-line structural(yaml-like) representation.
    a = {
    1: 3,
    'x': {1:4, 2:5},
    'l': [1, 2, 3],
    }
    for l in struct_repr(a):

        print l
    """
    # Output:
    # 1 : 3
    # l : - 1
    #     - 2
    #     - 3
    # x : 1 : 4
    # 2 : 5

    if type(data) in listtype:
        if len(data) == 0:
            return ["[]"]

        max_width = 0
        elt_lines = []
        for elt in data:
            sublines = struct_repr(elt)
            sublines_max_width = max([len(x) for x in sublines])

            if max_width < sublines_max_width:
                max_width = sublines_max_width

            elt_lines.append(sublines)

        lines = []
        for sublines in elt_lines:
            # - subline[0]
            #   subline[1]
            #   ...

            lines.append("- " + sublines[0].ljust(max_width))

            for line in sublines[1:]:
                lines.append("  " + line.ljust(max_width))

        return lines

    elif type(data) is dict:
        if len(data) == 0:
            return ["{}"]

        max_k_width = 0
        max_v_width = 0

        kvs = []

        for k, v in data.items():
            k = utf8str(k)
            sublines = struct_repr(v)
            sublines_max_width = max([len(x) for x in sublines])

            if max_k_width < len(k):
                max_k_width = len(k)

            if max_v_width < sublines_max_width:
                max_v_width = sublines_max_width

            kvs.append((k, sublines))

        kvs.sort(key=key)

        lines = []
        for k, sublines in kvs:
            # foo : sub-0
            #       sub-1
            #   b : sub-0
            #       sub-0

            lines.append(k.rjust(max_k_width) + " : " + sublines[0].ljust(max_v_width))

            for line in sublines[1:]:
                lines.append(" ".rjust(max_k_width) + "   " + line.ljust(max_v_width))

        return lines

    else:
        data = filter_invisible_chars(data)
        return [utf8str(data)]


def filter_invisible_chars(data):
    """
    Filters invisible characters in a string or a unicode object
    :param data: a string or unicode object to filter invisible characters
    :return: a filtered string or unicode object
    """
    # from pykit.strutil import filter_invisible_chars
    # cases = [
    #     "1273883926293937729\000\001\031",
    #     "\x00\x01\x02\x03\x04\005",
    #     u"1122299299299299292",
    #     u"\x00\x01\x02\x03\x04\005",
    # ]
    #
    # rst = []
    # for case in cases:
    #     rst.append(strutil.filter_invisible_chars(case))
    #
    # for r in rst:
    #     print(r)
    # '1273883926293937729'
    # ''
    # u'1122299299299299292'
    # u''
    if type(data) not in (bytes, str):
        return data

    return invisible_chars_re.sub("", data)


def _get_key_and_headers(keys, rows):
    if keys is None:
        if len(rows) == 0:
            keys = []
        else:
            r0 = rows[0]

            if type(r0) is dict:
                keys = list(r0.keys())
                keys.sort()
            elif type(r0) in listtype:
                keys = [i for i in range(len(r0))]
            else:
                keys = [""]

    _keys = []
    column_headers = []

    for k in keys:
        if type(k) not in listtype:
            k = [k, k]

        _keys.append(k[0])
        column_headers.append(str(k[1]))

    return _keys, column_headers


def utf8str(s):
    if isinstance(s, bytes):
        return str(s, "utf-8")
    return str(s)


def format_line(items, sep=" ", aligns=""):
    """
    It formats a list in a multi row manner.
    It is compatible with colored string such as those created with `strutil.blue("blue-text")`.
    :param items: elements in a line.
    Each element could be a `string` or a `list` of `string`.
    If it is a `list` of `string`, it would be rendered as a multi-row
    element.
    :param sep: specifies the separator between each element in a line.
    By default it is a single space `" "`.
    :param aligns: specifies alignment for each element.
    -   `l` for left-align.
    -   `r` for right-align.
    If no alignment specified for i-th element, it will be aligned to right by default.
    :return: formatted string.
    format a line with multi-row columns.
    """
    # items = [ 'name:',
    #           [ 'John',
    #             'j is my nick'],
    #           [ 'age:' ],
    #           [ 26, ],
    #           [ 'experience:' ],
    #           [ '2000 THU',
    #             '2006 sina',
    #             '2010 other'
    #             ],
    #           ]
    # format_line(items, sep=' | ', aligns = 'llllll')
    #
    # outputs:
    # name: | John         | age: | 26 | experience: | 2000 THU
    # | j is my nick |      |    |             | 2006 sina
    # |              |      |    |             | 2010 other

    aligns = [x for x in aligns] + [""] * len(items)
    aligns = aligns[: len(items)]
    aligns = ["r" if x == "r" else x for x in aligns]

    items = [(x if type(x) in listtype else [x]) for x in items]

    items = [[_to_str(y) for y in x] for x in items]

    maxHeight = max([len(x) for x in items] + [0])

    def max_width(x):
        return max([y.__len__() for y in x] + [0])

    widths = [max_width(x) for x in items]

    items = [(x + [""] * maxHeight)[:maxHeight] for x in items]

    lines = []
    for i in range(maxHeight):
        line = []
        for j in range(len(items)):
            width = widths[j]
            elt = items[j][i]

            actualWidth = elt.__len__()
            elt = utf8str(elt)
            if actualWidth < width:
                padding = " " * (width - actualWidth)
                if aligns[j] == "l":
                    elt = elt + padding
                else:
                    elt = padding + elt

            line.append(elt)

        line = sep.join(line)

        lines.append(line)

    return "\n".join(lines)


def format_table(rows, keys=None, colors=None, sep=" | ", row_sep=None):
    """
    Render a list of data into a table.
    Number of rows is `len(rows)`.
    Number of columns is `len(rows[0])`.
    :param rows: list of items to render.
    Element of list can be number, string, list or dict.
    :param keys: specifies indexes(for list) or keys(for dict) to render.
    It is a list.
    Indexes or keys those are not in this list will not be rendered.
    It can also be used to specify customized column headers, if element in
    list is a 2-element tuple or list:
    :param colors: specifies the color for each column.
    It is a list of color values in number or color name strings.
    If length of `colors` is smaller than the number of columns(the number of
    indexes of a list, or keys of a dict), the colors are repeated for columns
    after.
    :param sep: specifies char to separate rows.
    By default it is None, it means do not add line separator.
    :param row_sep: specifies column separator char.
    By default it is `" | "`.
    :return: a list of string.
    """
    keys, column_headers = _get_key_and_headers(keys, rows)
    colors = _get_colors(colors, len(keys))

    # element of lns is a mulit-column line
    # lns = [
    #         # line 1
    #         [
    #                 # column 1 of line 1
    #                 ['name:', # row 1 of column 1 of line 1
    #                  'foo',   # row 2 of column 1 of line 1
    #                 ],
    #
    #                 # column 2 of line 1
    #                 ['school:',
    #                  'foo',
    #                  'bar',
    #                 ],
    #         ],
    # ]

    # headers
    lns = [[[a + ": "] for a in column_headers]]

    for row in rows:
        if row_sep is not None:
            lns.append([[None] for k in keys])

        if type(row) is dict:
            ln = [struct_repr(row.get(k, "")) for k in keys]

        elif type(row) in listtype:
            ln = [struct_repr(row[int(k)]) if len(row) > int(k) else "" for k in keys]

        else:
            ln = [struct_repr(row)]

        lns.append(ln)

    def get_max_width(cols):
        return max([len(utf8str(c[0])) for c in cols] + [0])

    max_widths = [get_max_width(cols) for cols in zip(*lns)]

    rows = []
    for row in lns:
        ln = []

        for i in range(len(max_widths)):
            color = colors[i]
            w = max_widths[i]

            ln.append([k3color.Str(x.ljust(w), color) if x is not None else row_sep * w for x in row[i]])

        rows.append(format_line(ln, sep=sep))

    return rows


def _get_colors(colors, col_n):
    if colors is None:
        colors = []

    colors = colors or ([None] * col_n)

    while len(colors) < col_n:
        colors.extend(colors)

    colors = colors[:col_n]

    return colors


def _findquote(line, quote):
    if len(quote) == 0:
        return -1, -1, []

    i = 0
    n = len(line)
    escape = []
    while i < n:
        if line[i] == "\\":
            escape.append(i)
            i += 2
            continue

        if line[i] in quote:
            quote_s = i - len(escape)

            j = i
            i += 1
            while i < n and line[i] != line[j]:
                if line[i] == "\\":
                    escape.append(i)
                    i += 2
                    continue

                i += 1

            if i < n:
                quote_e = i - len(escape)
                return quote_s, quote_e, escape
            else:
                return quote_s, -1, escape

        i += 1

    return -1, -1, escape


def tokenize(line, sep=None, quote="\"'", preserve=False):
    """
    :param line: the line to tokenize.
    :param sep: is None or a non-empty string separator to tokenize with.
    If sep is None, runs of consecutive whitespace are regarded as a single
    separator, and the result will contain no empty strings at the start or end
    if the string has leading or trailing whitespace. Consequently, splitting
    an empty string or a string consisting of just whitespace with a None
    separator returns `[]`. Just like `str.split(None)`.
    By default, `sep` is None.
    :param quote:Every character in `quote` is regarded as a quote. Add a `\` prefix to make
    an exception. Segment between the same quotes is preserved.
    By default, `quote` is `'"\''`.

    :param preserve: preserve the quote itself if `preserve` is `True`.
    By default, `preserve` is `False`.
    :return: a list of string.
    """

    if sep == quote:
        raise ValueError("diffrent sep and quote is required")

    if sep is None:
        if len(line) == 0:
            return []
        line = line.strip()

    rst = [""]
    n = len(line)
    i = 0
    while i < n:
        quote_s, quote_e, escape = _findquote(line[i:], quote)

        if len(escape) > 0:
            lines = []
            x = 0
            for e in escape:
                lines.append(line[x : i + e])
                x = i + e + 1
            lines.append(line[x:])
            line = "".join(lines)
            n = len(line)

        if quote_s < 0:
            sub = n
        else:
            sub = i + quote_s

        if i < sub:
            sub_rst = line[i:sub].split(sep)
            if sep is None:
                if line[sub - 1] in string.whitespace:
                    sub_rst.append("")
                if line[i] in string.whitespace:
                    sub_rst.insert(0, "")

            head = rst.pop()
            sub_rst[0] = head + sub_rst[0]
            rst += sub_rst

        if quote_s < 0:
            break

        # discard incomplete
        # 'a b"c'  ->  ['a']
        if quote_e < 0:
            rst.pop()
            break

        head = rst.pop()

        if preserve:
            head += line[i + quote_s : i + quote_e + 1]
        else:
            head += line[i + quote_s + 1 : i + quote_e]

        rst.append(head)
        i += quote_e + 1

    return rst


def parse_colon_kvs(data):
    data = tokenize(data, quote="\"'")

    ret = {}
    for buf in data:
        if ":" not in buf:
            raise ValueError('invalid arguments, argumentsneed key-val like: "k:v"')

        k, v = buf.split(":", 1)

        ret[k] = v

    return ret


def page(lines, max_lines=10, control_char=True, pager=("less",)):
    """
    Display `lines` of string in console, with a pager program (`less`) if too many
    lines.

    It could be used in a interactive tool to display large content.

    It output strings directly to stdout.
    :param lines: is `list` of lines to display.
    :param max_lines: specifies the max lines not to use a pager.
    By default it is 10 lines.
    :param control_char: specifies if to interpret controlling chars, such as color char in terminal.
    :param pager: specifies the program as a pager.
    It is a list of command and argument.
    By default it is `('less',)`.
    :return: Nothing
    """
    if len(lines) > max_lines:
        pp = {"stdin": subprocess.PIPE, "stdout": None, "stderr": None}

        cmd_pager = list(pager)
        if control_char:
            if pager == ("less",):
                cmd_pager += ["-r"]

        subproc = subprocess.Popen(cmd_pager, close_fds=True, cwd="./", **pp)

        try:
            out, err = subproc.communicate(bytes("\n".join(lines).encode("utf-8")))
        except IOError as e:
            if e[0] == errno.EPIPE:
                pass
            else:
                raise
        subproc.wait()
    else:
        os.write(1, bytes(("\n".join(lines) + "\n").encode("utf-8")))



"""
#   Name

k3fmt

It provides with several string operation functions.

#   Status

This library is considered production ready.

"""

# from .proc import CalledProcessError
# from .proc import ProcError

from importlib.metadata import version

__version__ = version("k3fmt")

from .strutil import (
    format_line,
    break_line,
    parse_colon_kvs,
    tokenize,
    line_pad,
    struct_repr,
    format_table,
    filter_invisible_chars,
    utf8str,
    page,
)


__all__ = [
    "format_line",
    "parse_colon_kvs",
    "tokenize",
    "line_pad",
    "break_line",
    "struct_repr",
    "format_table",
    "filter_invisible_chars",
    "utf8str",
    "page",
]



import os
import unittest
import k3ut
import k3fmt
import k3color
import k3proc

dd = k3ut.dd

this_base = os.path.dirname(__file__)


def dd_lines(lines):
    dd()
    for line in lines:
        dd(line)


class TestK3fmt(unittest.TestCase):
    def test_format_line(self):
        cases = (
            (([], "", ""), "", ""),
            ((["a"], "", ""), "a", ""),
            (([["a", "bc"]], "", ""), " a\nbc", "default alignment is to right"),
            (([["a", "bc"], "foo"], "", ""), " afoo\nbc   ", "no sep, add space to align"),
            (([["a", "bc"], "foo"], "|", ""), " a|foo\nbc|   ", 'sep is "|"'),
            (
                ([["a", "bc"], "foo", [1, 2, 333]], "|", ""),
                " a|foo|  1\nbc|   |  2\n  |   |333",
                "number will be convert to str",
            ),
            (
                ([["a", "bc"], "foo", [1, 2, k3color.blue("xp")]], "|", ""),
                " a|foo| 1\nbc|   | 2\n  |   |" + str(k3color.blue("xp")),
                "k3color.Str instance is compatible",
            ),
        )

        for _in, _out, _mes in cases:
            rst = k3fmt.format_line(*_in)

            dd("_in: " + str(_in))
            dd("rst:\n" + rst)

            self.assertEqual(
                _out,
                rst,
                ("input: {_in}, output: {_out}, expected: {rst}, message: {_mes}").format(
                    _in=repr(_in), _out=repr(_out), rst=repr(rst), _mes=_mes
                ),
            )

    def test_format_table(self):
        inp = [
            {
                "acl": {},
                "bucket": "game1.read",
                "bucket_id": "1400000000000689036\000",
                "num_used": "0",
                "owner": "game1",
                "space_used": "0",
                "ts": "1492091893065708032",
            },
            {
                "acl": {},
                "bucket": "game2.read",
                "bucket_id": "1510000000000689037",
                "num_used": "0",
                "owner": "game2",
                "space_used": "0",
                "ts": "1492091906629786880",
            },
            {
                "acl": {"imgx": ["READ", "READ_ACP", "WRITE", "WRITE_ACP"]},
                "bucket": "imgx-test",
                "bucket_id": "1910000000000689048",
                "num_used": "0",
                "owner": "imgx",
                "space_used": "0",
                "ts": "1492101189213795840",
            },
        ]

        # default
        rst = k3fmt.format_table(inp)
        dd_lines(rst)

        # the last line is a '\n' splitted multi-row line
        expected = [
            "acl:               | bucket:    | bucket_id:          | num_used:  | owner:  | space_used:  | ts:                ",
            "{}                 | game1.read | 1400000000000689036 | 0          | game1   | 0            | 1492091893065708032",
            "{}                 | game2.read | 1510000000000689037 | 0          | game2   | 0            | 1492091906629786880",
            "imgx : - READ      | imgx-test  | 1910000000000689048 | 0          | imgx    | 0            | 1492101189213795840\n"
            "       - READ_ACP  |            |                     |            |         |              |                    \n"
            "       - WRITE     |            |                     |            |         |              |                    \n"
            "       - WRITE_ACP |            |                     |            |         |              |                    ",
        ]
        self.assertEqual(expected, rst)

        # specify key to render

        rst = k3fmt.format_table(inp, keys=[["bucket", "B"]])
        dd_lines(rst)
        expected = [
            "B:        ",
            "game1.read",
            "game2.read",
            "imgx-test ",
        ]
        self.assertEqual(expected, rst)

        # row_sep
        rst = k3fmt.format_table(inp, keys=["bucket"], row_sep="+")
        dd_lines(rst)
        expected = [
            "bucket:   ",
            "++++++++++",
            "game1.read",
            "++++++++++",
            "game2.read",
            "++++++++++",
            "imgx-test ",
        ]
        self.assertEqual(expected, rst)

        # sep
        rst = k3fmt.format_table(inp, keys=["bucket", "bucket_id"], sep=" # ")
        dd_lines(rst)
        expected = [
            "bucket:    # bucket_id:         ",
            "game1.read # 1400000000000689036",
            "game2.read # 1510000000000689037",
            "imgx-test  # 1910000000000689048",
        ]
        self.assertEqual(expected, rst)

    def test_struct_repr(self):
        inp = {
            1: 3,
            "x": {1: 4, 2: 5},
            "yyy": [1, 2, 3, 1000],
        }

        rst = k3fmt.struct_repr(inp)

        for line in rst:
            dd(repr(line))

        expected = [
            "  1 : 3     ",
            "  x : 1 : 4 ",
            "      2 : 5 ",
            "yyy : - 1   ",
            "      - 2   ",
            "      - 3   ",
            "      - 1000",
        ]

        self.assertEqual(expected, rst)

    def test_utf8str(self):
        cases = (
            ("", ""),
            ("1", "1"),
            (1, "1"),
            (b"\xe6\x88\x91", "我"),
            ("\xe6\x88\x91", "\xe6\x88\x91"),
        )

        for inp, expected in cases:
            rst = k3fmt.utf8str(inp)

            self.assertEqual(expected, rst)

    def test_filter_invisible_chars(self):
        cases = (
            ("1273883926293937729\000\001\031", "1273883926293937729"),
            ("\x00\x01\x02\x03\x04\005", ""),
            ("1122299299299299292", "1122299299299299292"),
            ("\x00\x01\x02\x03\x04\005", ""),
            (None, None),
            ("", ""),
        )

        for case, expected in cases:
            self.assertEqual(expected, k3fmt.filter_invisible_chars(case))

    def test_break_line(self):
        case = [
            ("a quick brown fox jumps over the lazy dog", 13, ["a quick brown", "fox jumps", "over the lazy", "dog"]),
            ("a quick\nbrown fox jumps over the lazy dog", 13, ["a quick", "brown fox", "jumps over", "the lazy dog"]),
            ("a quick\rbrown fox jumps over the lazy dog", 13, ["a quick", "brown fox", "jumps over", "the lazy dog"]),
            (
                "a quick brown fox jumps\r\nover the lazy dog",
                13,
                ["a quick brown", "fox jumps", "over the lazy", "dog"],
            ),
            (
                "a quick\nbrown\rfox jumps\r\nover the lazy dog",
                13,
                ["a quick", "brown", "fox jumps", "over the lazy", "dog"],
            ),
            ("aquickbrownfoxjumpsoverthelazydog", 9, ["aquickbrownfoxjumpsoverthelazydog"]),
            ("aquickbro", 9, ["aquickbro"]),
            (" aquickbro", 9, ["", "aquickbro"]),
            ("  aquickbro", 9, [" ", "aquickbro"]),
            ("aquickbro ", 9, ["aquickbro"]),
            ("aquickbro  ", 9, ["aquickbro", " "]),
            ("aqu ick br", 9, ["aqu ick", "br"]),
            ("aqu ick br", 9.34, ["aqu ick", "br"]),
            ("apu   ick  br", 5, ["apu  ", "ick ", "br"]),
            ("aqu ick br", 0, ["aqu", "ick", "br"]),
            ("aqu ick br", -1, ["aqu", "ick", "br"]),
            ("", 2, []),
            (" ", 2, [" "]),
            ("  ", 2, ["  "]),
            ("   ", 2, ["  "]),
            ("    ", 2, ["  ", " "]),
        ]

        for linestr, width, expected in case:
            rst = k3fmt.break_line(linestr, width)
            dd("case: ", linestr, width, expected)
            dd("rst: ", rst)
            self.assertEqual(rst, expected)

        colored_string_cases = [
            (
                [("asd ", "red"), ("fer ", "blue"), ("fg", "white")],
                8,
                [[("asd", "red"), (" ", None), ("fer", "blue")], [("fg", "white")]],
                "base test",
            ),
            (
                [("asd ", "red"), ("fer ", "blue"), ("fg", "white")],
                -1,
                [[("asd", "red")], [("fer", "blue")], [("fg", "white")]],
                "test width",
            ),
            (
                [("asd   ", "red"), ("fer  ", "blue"), ("fg", "white")],
                5,
                [[("asd", "red"), (" ", None), (" ", None)], [("fer", "blue"), (" ", None)], [("fg", "white")]],
                "test consecutive blank spaces",
            ),
            (
                [("  asd", "red"), (" fer", "blue"), ("fg   ", "white")],
                4,
                [[(" ", None)], [("asd", "red")], [("fer", "blue"), ("fg", "white")], [(" ", None), (" ", None)]],
                "consecutive blank spaces in the end and beginning",
            ),
            (
                [("asd\n", "red"), ("f er ", "blue"), ("fg", "white")],
                5,
                [[("asd", "red")], [("f", "blue"), (" ", None), ("er", "blue")], [("fg", "white")]],
                "test line break",
            ),
            (
                [("", None)],
                2,
                [],
                "test empty string",
            ),
        ]

        for _in, width, expected, msg in colored_string_cases:
            dd("msg: ", msg)

            color_in = k3color.Str("")
            for elt in _in:
                color_in += k3color.Str(elt[0], elt[1])
            color_expected = []
            for exp_line in expected:
                buf = k3color.Str("")
                for elt in exp_line:
                    buf += k3color.Str(elt[0], elt[1])
                color_expected.append(buf)

            rst = k3fmt.break_line(color_in, width)

            dd("rst: ", rst)
            dd("expected: ", color_expected)
            self.assertEqual(rst, color_expected)

    def test_line_pad(self):
        cases = (
            (("", ""), "", ""),
            (("a", ""), "a", ""),
            (("a", " "), " a", ""),
            (("a", "   "), "   a", ""),
            (("a\nb", "   "), "   a\n   b", ""),
            (("a\nbc", lambda line: "x" * len(line)), "xa\nxxbc", ""),
        )

        for _in, _out, _mes in cases:
            rst = k3fmt.line_pad(*_in)
            self.assertEqual(
                _out,
                rst,
                ("input: {_in}, output: {_out}, expected: {rst}, message: {_mes}").format(
                    _in=repr(_in), _out=repr(_out), rst=repr(rst), _mes=_mes
                ),
            )

    def test_tokenize(self):
        base_cases = (
            ("", [""], ""),
            ("a", ["a"], ""),
            ("a b", ["a", "b"], ""),
            (" a", ["", "a"], ""),
            ("b ", ["b", ""], ""),
            ("abc def gh", ["abc", "def", "gh"], ""),
            ('"ab cd"', ['"ab cd"'], ""),
            ('"ab  cd"', ['"ab  cd"'], "multiple space inside quotes"),
            ('"ab cd" ', ['"ab cd"', ""], ""),
            (' "ab cd"', ["", '"ab cd"'], ""),
            ('"ab cd" "x', ['"ab cd"'], "discard incomplete"),
            ('"ab cd" "x y"', ['"ab cd"', '"x y"'], ""),
            ('foo "ab cd" "x y"', ["foo", '"ab cd"', '"x y"'], ""),
            ('foo "ab cd" "x', ["foo", '"ab cd"'], "discard incomplete"),
            ('foo "\\"ab cd" "x', ["foo", '""ab cd"'], 'escape "'),
            ('foo "\\\\"ab cd" "x', ["foo", '"\\"ab', 'cd" "x'], "escape \\"),
            ('foo "\\\\\\"ab cd" "x', ["foo", '"\\"ab cd"'], 'escape \\ "'),
            ('a \\"bc "d e" "f', ["a", '"bc', '"d e"'], ""),
            ('a \\\\"bc "d e" "f', ["a", '\\"bc "d', 'e" "f'], ""),
            ('a \\\\\\"bc "d e" "f', ["a", '\\"bc', '"d e"'], ""),
            ('a "bc "d \\"f', ["a", '"bc "d', '"f'], ""),
            ('a "bc "d \\\\"f', ["a", '"bc "d'], ""),
            ('a "bc "d \\\\\\"f', ["a", '"bc "d', '\\"f'], ""),
            ('\\"bc "d "f', ['"bc', '"d "f'], ""),
            ('\\\\"bc "d "f', ['\\"bc "d'], ""),
            ('\\\\\\"bc "d "f', ['\\"bc', '"d "f'], ""),
            ('a "bc "d f\\"', ["a", '"bc "d', 'f"'], ""),
            ('a "bc "d f\\\\"', ["a", '"bc "d'], ""),
            ('a "bc "d f\\\\\\"', ["a", '"bc "d', 'f\\"'], ""),
        )

        for _in, _out, _mes in base_cases:
            rst = k3fmt.tokenize(_in, sep=" ", preserve=True)
            self.assertEqual(
                _out,
                rst,
                ("input: {_in}, output: {_out}, expected: {rst}, message: {_mes}").format(
                    _in=repr(_in), _out=repr(_out), rst=repr(rst), _mes=_mes
                ),
            )

        sep_cases = (
            ("", None, True, []),
            (" a  b  c ", None, True, ["a", "b", "c"]),
            (' a  "b  c" ', None, True, ["a", '"b  c"']),
            (' a  "b  c" ', None, False, ["a", "b  c"]),
            ("a b c", None, True, ["a", "b", "c"]),
            ('"a b c"', None, True, ['"a b c"']),
            ('"a b c"', None, False, ["a b c"]),
            ('a b"c d"', None, True, ["a", 'b"c d"']),
            ('a b"c d"', None, False, ["a", "bc d"]),
            ("a bcd", "bc", True, ["a ", "d"]),
            ('a "bc" d', "bc", True, ['a "bc" d']),
            ('a "bc" d', "bc", False, ["a bc d"]),
            ("abcd", "abcd", True, ["", ""]),
        )

        for line, sep, preserve, rst_expected in sep_cases:
            dd("in: ", line, sep)
            rst = k3fmt.tokenize(line, sep=sep, quote='"', preserve=preserve)
            dd("out: ", rst)
            self.assertEqual(rst, rst_expected)

        preserve_cases = (
            ('""', '"', True, ['""']),
            ('""', '"', False, [""]),
            ("abc xd efx gh", "x", True, ["abc", "xd efx", "gh"]),
            ("abc xd efx gh", "x", False, ["abc", "d ef", "gh"]),
            ("ab cxd efx gh", "x", True, ["ab", "cxd efx", "gh"]),
            ("ab cxd efx gh", "x", False, ["ab", "cd ef", "gh"]),
            ("ab cxd efxgh", "x", True, ["ab", "cxd efxgh"]),
            ("ab cxd efxgh", "x", False, ["ab", "cd efgh"]),
            ("ab cxd yey fx gh", "xy", True, ["ab", "cxd yey fx", "gh"]),
            ("ab cxd yey fx gh", "xy", False, ["ab", "cd yey f", "gh"]),
            ("ab cxd yey f gh", "xy", True, ["ab"]),
            ("ab cxd yey f gh", "xy", False, ["ab"]),
            ("ab cxd xex f gh", "x", True, ["ab"]),
            ("ab cxd xex f gh", "x", False, ["ab"]),
        )

        for line, quote, preserve, rst_expected in preserve_cases:
            dd("in: ", line, quote, preserve)
            rst = k3fmt.tokenize(line, sep=" ", quote=quote, preserve=preserve)
            dd("out: ", rst)
            self.assertEqual(rst, rst_expected)

    def test_parse_colon_kvs(self):
        cases = (
            ("abc:123", {"abc": "123"}),
            (" abc:123", {"abc": "123"}),
            ("abc:123 ", {"abc": "123"}),
            (" abc:123 ", {"abc": "123"}),
            ("    abc:123", {"abc": "123"}),
            ("abc:123   ", {"abc": "123"}),
            ("    abc:123   ", {"abc": "123"}),
            (" 'a:' bc:123", {"a": "", "bc": "123"}),
            ("abc: '123:' ", {"abc": "", "123": ""}),
            ("a:bc:123", {"a": "bc:123"}),
            ("abc:1:23", {"abc": "1:23"}),
            ("abc:\n123:", {"abc": "", "123": ""}),
            ("abc:\n123:\t", {"abc": "", "123": ""}),
            ("abc:\n\t123:", {"abc": "", "123": ""}),
            ("abc:123 abc:456", {"abc": "456"}),
            ("abc:123 def:456", {"abc": "123", "def": "456"}),
            ("", {}),
            (":", {"": ""}),
            ("::", {"": ":"}),
        )

        for case, exp in cases:
            self.assertEqual(k3fmt.parse_colon_kvs(case), exp)

    def test_raise_value_error(self):
        cases = (
            "abc123",
            "abc:1 23",
            "abc: 123",
            "abc:   123",
        )

        for case in cases:
            self.assertRaises(ValueError, k3fmt.parse_colon_kvs, case)

    def test_page_no_pager(self):
        returncode, out, err = k3proc.command(
            "python",
            "page_it.py",
            # pager, control_char, max_lines
            " ".join(["python", "raw_pager.py", ">"]),
            "0",
            "2",
            "1",
            "2",
            cwd=this_base,
        )
        dd("returncode:", returncode)
        dd("out:", out)
        dd("err:", err)

        self.assertEqual(0, returncode)
        self.assertEqual("1\n2\n", out)

    def test_page_use_pager(self):
        returncode, out, err = k3proc.command(
            "python",
            "page_it.py",
            # pager, control_char, max_lines
            " ".join(["python", "raw_pager.py", ">"]),
            "0",
            "2",
            "1",
            "2",
            "3",
            cwd=this_base,
        )
        dd("returncode:", returncode)
        dd("out:", out)
        dd("err:", err)

        self.assertEqual(0, returncode)
        self.assertEqual("> 1\n> 2\n> 3\n", out)



import k3fmt

lines = [
    "hello",
    "world",
]

# add left padding to each line in a string
k3fmt.line_pad("\n".join(lines), " " * 4)
# "    hello"
# "    world"


# format a multi-row line
items = [
    "name:",
    ["John", "j is my nick"],
    "age:",
    26,
    "experience:",
    ["2000 THU", "2006 sina", "2010 other"],
]

k3fmt.format_line(items, sep=" | ", aligns="llllll")
# outputs:
#    name: | John         | age: | 26 | experience: | 2000 THU
#          | j is my nick |      |    |             | 2006 sina
#          |              |      |    |             | 2010 other


# ['ab']
print(k3fmt.tokenize("ab"))
# ['a', 'b']
print(k3fmt.tokenize("a b"))
# ['a', 'b']
print(k3fmt.tokenize(" a  b "))
# ['a', 'b']
print(k3fmt.tokenize(" a\t b\n c\r "))
# ['a b', 'c d']
print(k3fmt.tokenize("a bxyc d", sep="xy"))
# ['a', 'x x', 'b']
print(k3fmt.tokenize('a "x x" b'))
# ['a', 'x x', 'b']
print(k3fmt.tokenize("a 'x x' b 'x"))  # the last `'x` has no pair, discard
# ['a', 'a b', 'c d']
print(k3fmt.tokenize(" a  xa bx yc dy ", quote="xy"))
# ['a', 'xa bx', 'yc dy']
print(k3fmt.tokenize("a xa bx yc dy", quote="xy", preserve=True))
# ['', 'a', 'xa bx', 'yc dy', '']
print(k3fmt.tokenize(" a xa bx yc dy ", sep=" ", quote="xy", preserve=True))
